"""aws_cdk_constructs package."""
from .microservice import Microservice
from .database import Database
from .bucket import Bucket
from .api import API
from .service_user_for_iac import ServiceUserForIAC
from .service_user_for_static_assets import ServiceUserForStaticAssets


__version__ = '2.1.1'
